# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Type, TypeVar, Any, Union, Optional, Tuple, Iterator, Callable, NewType
import attr
import copy
import sys

from ..primitive import JSON
from .serializable import SerializerError, Serializable, GenericSerializable
from .obj import Obj, Lst

T = TypeVar("T")
T2 = TypeVar("T2")

Serializer = NewType("Serializer", Callable[[T], JSON])
Deserializer = NewType("Deserializer", Callable[[JSON], T])
serializer_map: Dict[Type[T], Serializer] = {}
deserializer_map: Dict[Type[T], Deserializer] = {}

no_value = object()


def serializer(elem_type: Type[T]) -> Callable[[Serializer], Serializer]:
    """
    Define a custom serialization function for the given type.

    Args:
        elem_type: The type to define the serializer for.

    Returns:
        Decorator for the function.

    Examples:
        >>> from datetime import datetime
        >>> from mautrix.types import serializer, JSON
        >>> @serializer(datetime)
        ... def serialize_datetime(dt: datetime) -> JSON:
        ...     return dt.timestamp()
    """

    def decorator(func: Serializer) -> Serializer:
        serializer_map[elem_type] = func
        return func

    return decorator


def deserializer(elem_type: Type[T]) -> Callable[[Deserializer], Deserializer]:
    """
    Define a custom deserialization function for a given type hint.

    Args:
        elem_type: The type hint to define the deserializer for.

    Returns:
        Decorator for the function:

    Examples:
        >>> from datetime import datetime
        >>> from mautrix.types import deserializer, JSON
        >>> @deserializer(datetime)
        ... def deserialize_datetime(data: JSON) -> datetime:
        ...     return datetime.fromtimestamp(data)
    """

    def decorator(func: Deserializer) -> Deserializer:
        deserializer_map[elem_type] = func
        return func

    return decorator


def _fields(attrs_type: Type[T], only_if_flatten: bool = None) -> Iterator[Tuple[str, Type[T2]]]:
    return ((field.metadata.get("json", field.name), field) for field in attr.fields(attrs_type)
            if only_if_flatten is None or field.metadata.get("flatten", False) == only_if_flatten
            and not field.metadata.get("hidden", False))


immutable = (int, str, float, bool, type(None))


def _safe_default(val: T) -> T:
    if isinstance(val, immutable):
        return val
    return copy.copy(val)


def _dict_to_attrs(attrs_type: Type[T], data: JSON, default: Optional[T] = None,
                   default_if_empty: bool = False) -> T:
    data = data or {}
    unrecognized = {}
    new_items = {field.name.lstrip("_"):
                     _try_deserialize(field.type, data, field.default,
                                      field.metadata.get("ignore_errors", False))
                 for _, field in _fields(attrs_type, only_if_flatten=True)}
    fields = dict(_fields(attrs_type, only_if_flatten=False))
    for key, value in data.items():
        try:
            field = fields[key]
        except KeyError:
            unrecognized[key] = value
            continue
        name = field.name.lstrip("_")
        new_items[name] = _try_deserialize(field.type, value, field.default,
                                           field.metadata.get("ignore_errors", False))
    if len(new_items) == 0 and default_if_empty and default is not attr.NOTHING:
        return _safe_default(default)
    try:
        obj = attrs_type(**new_items)
    except TypeError as e:
        for key, field in _fields(attrs_type):
            json_key = field.metadata.get("json", key)
            if field.default is attr.NOTHING and json_key not in new_items:
                raise SerializerError(
                    f"Missing value for required key {field.name} in {attrs_type.__name__}") from e
        raise SerializerError("Unknown serialization error") from e
    if len(unrecognized) > 0:
        obj.unrecognized_ = unrecognized
    return obj


def _try_deserialize(cls: Type[T], value: JSON, default: Optional[T] = None,
                     ignore_errors: bool = False) -> T:
    try:
        return _deserialize(cls, value, default)
    except SerializerError:
        if not ignore_errors:
            raise
    except (TypeError, ValueError, KeyError) as e:
        if not ignore_errors:
            raise SerializerError("Unknown serialization error") from e


def _has_custom_deserializer(cls) -> bool:
    return (issubclass(cls, Serializable) and
            getattr(cls.deserialize, '__func__')
            != getattr(SerializableAttrs.deserialize, '__func__'))


def _deserialize(cls: Type[T], value: JSON, default: Optional[T] = None) -> T:
    if value is None:
        return _safe_default(default)

    try:
        deser = deserializer_map[cls]
    except KeyError:
        pass
    else:
        return deser(value)
    supertype = getattr(cls, "__supertype__", None)
    if supertype:
        cls = supertype
        try:
            deser = deserializer_map[supertype]
        except KeyError:
            pass
        else:
            return deser(value)
    if attr.has(cls):
        if _has_custom_deserializer(cls):
            return cls.deserialize(value)
        return _dict_to_attrs(cls, value, default, default_if_empty=True)
    elif cls == Any or cls == JSON:
        return value
    elif getattr(cls, "__origin__", None) is Union:
        if len(cls.__args__) == 2 and isinstance(None, cls.__args__[1]):
            return _deserialize(cls.__args__[0], value, default)
    elif isinstance(cls, type):
        if issubclass(cls, Serializable):
            return cls.deserialize(value)

    type_class = _get_type_class(cls)
    args = getattr(cls, "__args__", None)
    if type_class == list:
        item_cls, = args
        return [_deserialize(item_cls, item) for item in value]
    elif type_class == set:
        item_cls, = args
        return {_deserialize(item_cls, item) for item in value}
    elif type_class == dict:
        key_cls, val_cls = args
        return {_deserialize(key_cls, key): _deserialize(val_cls, item)
                for key, item in value.items()}

    if isinstance(value, list):
        return Lst(value)
    elif isinstance(value, dict):
        return Obj(**value)
    return value


if sys.version_info >= (3, 7):
    def _get_type_class(typ):
        try:
            return typ.__origin__
        except AttributeError:
            return None
else:
    def _get_type_class(typ):
        try:
            return typ.__extra__
        except AttributeError:
            return None


def _attrs_to_dict(data: T) -> JSON:
    new_dict = {}
    for json_name, field in _fields(data.__class__):
        if not json_name:
            continue
        field_val = getattr(data, field.name)
        if field_val is None:
            if not field.metadata.get("omitempty", True):
                field_val = field.default
            else:
                continue
        if field.metadata.get("omitdefault", False) and field_val == field.default:
            continue
        try:
            serialized = serializer_map[field.type](field_val)
        except KeyError:
            serialized = _serialize(field_val)
        if field.metadata.get("flatten", False) and isinstance(serialized, dict):
            new_dict.update(serialized)
        elif serialized != no_value:
            new_dict[json_name] = serialized
    try:
        new_dict.update(data.unrecognized_)
    except (AttributeError, TypeError):
        pass
    return new_dict


def _serialize(val: Any) -> JSON:
    if isinstance(val, Serializable):
        return val.serialize()
    elif isinstance(val, (tuple, list, set)):
        return [_serialize(subval) for subval in val]
    elif isinstance(val, dict):
        return {_serialize(subkey): _serialize(subval) for subkey, subval in val.items()}
    elif attr.has(val.__class__):
        return _attrs_to_dict(val)
    return val


class SerializableAttrs(GenericSerializable[T]):
    """
    An abstract :class:`Serializable` that assumes the subclass is an attrs dataclass.

    Examples:
        >>> from attr import dataclass
        >>> from mautrix.types import SerializableAttrs
        >>> @dataclass
        ... class Foo(SerializableAttrs['Foo']):
        ...     index: int
        ...     field: Optional[str] = None
    """
    unrecognized_: Optional[JSON]

    def __init__(self):
        self.unrecognized_ = {}

    @classmethod
    def deserialize(cls, data: JSON) -> T:
        return _dict_to_attrs(cls, data)

    def serialize(self) -> JSON:
        return _attrs_to_dict(self)

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            try:
                return self.unrecognized_[item]
            except AttributeError:
                self.unrecognized_ = {}
                raise KeyError(item)

    def __setitem__(self, item, value):
        if hasattr(self, item):
            setattr(self, item, value)
        else:
            try:
                self.unrecognized_[item] = value
            except AttributeError:
                self.unrecognized_ = {
                    item: value,
                }



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Awaitable, List, Tuple, Union, TYPE_CHECKING
from urllib.parse import quote as urllib_quote
import asyncio

from mautrix.api import Method, Path
from mautrix.types import (StateEvent, EventType, StateEventContent, EventID, ContentURI,
                           EventContent, UserID, RoomID, PresenceState,
                           RoomAvatarStateEventContent, RoomNameStateEventContent,
                           RoomTopicStateEventContent, PowerLevelStateEventContent,
                           RoomPinnedEventsStateEventContent, Membership, Member)
from mautrix.client import ClientAPI, StoreUpdatingAPI
from mautrix.errors import MForbidden, MBadState, MatrixRequestError, IntentError
from mautrix.util.logging import TraceLogger

from ..state_store import ASStateStore

try:
    import magic
except ImportError:
    magic = None

if TYPE_CHECKING:
    from .appservice import AppServiceAPI


def quote(*args, **kwargs):
    return urllib_quote(*args, **kwargs, safe="")


ENSURE_REGISTERED_METHODS = (
    # Room methods
    ClientAPI.create_room, ClientAPI.add_room_alias, ClientAPI.remove_room_alias,
    ClientAPI.get_room_alias, ClientAPI.get_joined_rooms, ClientAPI.join_room_by_id,
    ClientAPI.join_room, ClientAPI.set_room_directory_visibility, ClientAPI.forget_room,
    # User data methods
    ClientAPI.search_users, ClientAPI.set_displayname, ClientAPI.set_avatar_url,

    ClientAPI.upload_media,
)

ENSURE_JOINED_METHODS = (
    # Room methods
    ClientAPI.invite_user,
    # Event methods
    ClientAPI.get_event, ClientAPI.get_state_event, ClientAPI.get_state,
    ClientAPI.get_joined_members, ClientAPI.get_messages, ClientAPI.send_state_event,
    ClientAPI.send_message_event, ClientAPI.redact,
    ClientAPI.send_receipt, ClientAPI.set_fully_read_marker,
)


class IntentAPI(StoreUpdatingAPI):
    """
    IntentAPI is a high-level wrapper around the AppServiceAPI that provides many easy-to-use
    functions for accessing the client-server API. It is designed for appservices and will
    automatically handle many things like missing invites using the appservice bot.
    """
    api: 'AppServiceAPI'
    state_store: ASStateStore
    bot: 'IntentAPI'
    log: TraceLogger

    def __init__(self, mxid: UserID, api: 'AppServiceAPI', bot: 'IntentAPI' = None,
                 state_store: ASStateStore = None):
        super().__init__(mxid=mxid, api=api, state_store=state_store)
        self.bot = bot
        self.log = api.base_log.getChild("intent")

        for method in ENSURE_REGISTERED_METHODS:
            method = getattr(self, method.__name__)

            async def wrapper(*args, __self=self, __method=method, **kwargs):
                await __self.ensure_registered()
                return await __method(*args, **kwargs)

            setattr(self, method.__name__, wrapper)

        for method in ENSURE_JOINED_METHODS:
            method = getattr(self, method.__name__)

            async def wrapper(*args, __self=self, __method=method, **kwargs):
                room_id = kwargs.get("room_id", None)
                if not room_id:
                    room_id = args[0]
                await __self.ensure_joined(room_id)
                return await __method(*args, **kwargs)

            setattr(self, method.__name__, wrapper)

    def user(self, user_id: UserID, token: Optional[str] = None) -> 'IntentAPI':
        """
        Get the intent API for a specific user.
        This is just a proxy to :meth:`AppServiceAPI.intent`.

        You should only call this method for the bot user. Calling it with child intent APIs will
        result in a warning log.

        Args:
            user_id: The Matrix ID of the user whose intent API to get.
            token: The access token to use for the Matrix ID.

        Returns:
            The IntentAPI for the given user.
        """
        if not self.bot:
            return self.api.intent(user_id, token)
        else:
            self.log.warning("Called IntentAPI#user() of child intent object.")
            return self.bot.api.intent(user_id, token)

    # region User actions

    async def set_presence(self, presence: PresenceState = PresenceState.ONLINE,
                           status: Optional[str] = None, ignore_cache: bool = False):
        """
        Set the online status of the user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#put-matrix-client-r0-presence-userid-status>`__

        Args:
            presence: The online status of the user.
            status: The status message.
            ignore_cache: Whether or not to set presence even if the cache says the presence is
                already set to that value.
        """
        await self.ensure_registered()
        if not ignore_cache and self.state_store.has_presence(self.mxid, status):
            return
        await super().set_presence(presence, status)
        self.state_store.set_presence(self.mxid, status)

    # endregion
    # region Room actions

    async def invite_user(self, room_id: RoomID, user_id: UserID, check_cache: bool = False
                          ) -> None:
        """
        Invite a user to participate in a particular room. They do not start participating in the
        room until they actually join the room.

        Only users currently in the room can invite other users to join that room.

        If the user was invited to the room, the homeserver will add a ``m.room.member`` event to
        the room.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#post-matrix-client-r0-rooms-roomid-invite>`__

        Args:
            room_id: The ID of the room to which to invite the user.
            user_id: The fully qualified user ID of the invitee.
            check_cache: Whether or not the function should be a no-op if the state store says the
                user is already invited.
        """
        try:
            ok_states = (Membership.INVITE, Membership.JOIN)
            do_invite = (not check_cache or (await self.state_store.get_membership(room_id, user_id)
                                             not in ok_states))
            if do_invite:
                await super().invite_user(room_id, user_id)
                await self.state_store.invited(room_id, user_id)
        except MatrixRequestError as e:
            if e.errcode != "M_FORBIDDEN":
                raise IntentError(f"Failed to invite {user_id} to {room_id}", e)
            if "is already in the room" in e.message:
                await self.state_store.joined(room_id, user_id)

    def set_room_avatar(self, room_id: RoomID, avatar_url: Optional[ContentURI], **kwargs
                        ) -> Awaitable[EventID]:
        return self.send_state_event(room_id, EventType.ROOM_AVATAR,
                                     RoomAvatarStateEventContent(url=avatar_url),
                                     **kwargs)

    def set_room_name(self, room_id: RoomID, name: str, **kwargs) -> Awaitable[EventID]:
        return self.send_state_event(room_id, EventType.ROOM_NAME,
                                     RoomNameStateEventContent(name=name), **kwargs)

    def set_room_topic(self, room_id: RoomID, topic: str, **kwargs) -> Awaitable[EventID]:
        return self.send_state_event(room_id, EventType.ROOM_TOPIC,
                                     RoomTopicStateEventContent(topic=topic), **kwargs)

    async def get_power_levels(self, room_id: RoomID, ignore_cache: bool = False
                               ) -> PowerLevelStateEventContent:
        await self.ensure_joined(room_id)
        if not ignore_cache:
            levels = await self.state_store.get_power_levels(room_id)
            if levels:
                return levels
        levels = await self.get_state_event(room_id, EventType.ROOM_POWER_LEVELS)
        await self.state_store.set_power_levels(room_id, levels)
        return levels

    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent, **kwargs
                               ) -> EventID:
        response = await self.send_state_event(room_id, EventType.ROOM_POWER_LEVELS, content,
                                               **kwargs)
        if response:
            await self.state_store.set_power_levels(room_id, content)
        return response

    async def get_pinned_messages(self, room_id: RoomID) -> List[EventID]:
        await self.ensure_joined(room_id)
        content = await self.get_state_event(room_id, EventType.ROOM_PINNED_EVENTS)
        return content["pinned"]

    def set_pinned_messages(self, room_id: RoomID, events: List[EventID], **kwargs
                            ) -> Awaitable[EventID]:
        return self.send_state_event(room_id, EventType.ROOM_PINNED_EVENTS,
                                     RoomPinnedEventsStateEventContent(pinned=events), **kwargs)

    async def pin_message(self, room_id: RoomID, event_id: EventID) -> None:
        events = await self.get_pinned_messages(room_id)
        if event_id not in events:
            events.append(event_id)
            await self.set_pinned_messages(room_id, events)

    async def unpin_message(self, room_id: RoomID, event_id: EventID):
        events = await self.get_pinned_messages(room_id)
        if event_id in events:
            events.remove(event_id)
            await self.set_pinned_messages(room_id, events)

    async def set_join_rule(self, room_id: RoomID, join_rule: str, **kwargs):
        if join_rule not in ("public", "knock", "invite", "private"):
            raise ValueError(f"Invalid join rule \"{join_rule}\"")
        await self.send_state_event(room_id, EventType.ROOM_JOIN_RULES, {
            "join_rule": join_rule,
        }, **kwargs)

    async def get_room_displayname(self, room_id: RoomID, user_id: UserID, ignore_cache=False
                                   ) -> str:
        return (await self.get_room_member_info(room_id, user_id, ignore_cache)).displayname

    async def get_room_avatar_url(self, room_id: RoomID, user_id: UserID, ignore_cache=False
                                  ) -> str:
        return (await self.get_room_member_info(room_id, user_id, ignore_cache)).avatar_url

    async def get_room_member_info(self, room_id: RoomID, user_id: UserID, ignore_cache=False
                                   ) -> Member:
        member = await self.state_store.get_member(room_id, user_id)
        if not member or not member.membership or ignore_cache:
            member = await self.get_state_event(room_id, EventType.ROOM_MEMBER, user_id)
        return member

    async def set_typing(self, room_id: RoomID, is_typing: bool = True, timeout: int = 5000,
                         ignore_cache: bool = False) -> None:
        await self.ensure_joined(room_id)
        if not ignore_cache and is_typing == self.state_store.is_typing(room_id, self.mxid):
            return
        await super().set_typing(room_id, timeout if is_typing else 0)
        self.state_store.set_typing(room_id, self.mxid, is_typing, timeout)

    async def error_and_leave(self, room_id: RoomID, text: Optional[str] = None,
                              html: Optional[str] = None) -> None:
        await self.ensure_joined(room_id)
        await self.send_notice(room_id, text, html=html)
        await self.leave_room(room_id)

    async def send_message_event(self, room_id: RoomID, event_type: EventType,
                                 content: EventContent, **kwargs) -> EventID:
        await self._ensure_has_power_level_for(room_id, event_type)

        if self.api.is_real_user and self.api.real_user_content_key:
            content[self.api.real_user_content_key] = True

        return await super().send_message_event(room_id, event_type, content, **kwargs)

    async def send_state_event(self, room_id: RoomID, event_type: EventType,
                               content: Union[StateEventContent, Dict],
                               state_key: Optional[str] = "", **kwargs) -> EventID:
        await self._ensure_has_power_level_for(room_id, event_type)
        return await super().send_state_event(room_id, event_type, content, state_key, **kwargs)

    async def get_room_members(self, room_id: RoomID,
                               allowed_memberships: Tuple[Membership, ...] = (Membership.JOIN,)
                               ) -> List[UserID]:
        if len(allowed_memberships) == 1 and allowed_memberships[0] == Membership.JOIN:
            memberships = await self.get_joined_members(room_id)
            return list(memberships.keys())
        member_events = await self.get_members(room_id)
        return [evt.state_key for evt in member_events
                if evt.content.membership in allowed_memberships]

    async def leave_room(self, room_id: RoomID) -> None:
        if not room_id:
            raise ValueError("Room ID not given")
        await self.ensure_registered()
        try:
            await self.state_store.left(room_id, self.mxid)
            await super().leave_room(room_id)
        except MatrixRequestError as e:
            if "not in room" not in e.message:
                raise

    async def get_state(self, room_id: RoomID) -> List[StateEvent]:
        state = await super().get_state(room_id)
        await asyncio.gather(*[self.state_store.update_state(evt) for evt in state])
        return state

    async def mark_read(self, room_id: RoomID, event_id: EventID) -> None:
        if self.state_store.get_read(room_id, self.mxid) != event_id:
            await self.set_fully_read_marker(room_id, fully_read=event_id, read_receipt=event_id)
            self.state_store.set_read(room_id, self.mxid, event_id)

    # endregion
    # region Ensure functions

    async def ensure_joined(self, room_id: RoomID, ignore_cache: bool = False) -> None:
        if not room_id:
            raise ValueError("Room ID not given")
        if not ignore_cache and await self.state_store.is_joined(room_id, self.mxid):
            return
        await self.ensure_registered()
        try:
            await self.join_room(room_id, max_retries=0)
            await self.state_store.joined(room_id, self.mxid)
        except MForbidden as e:
            if not self.bot:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e
            try:
                await self.bot.invite_user(room_id, self.mxid)
                await self.join_room(room_id, max_retries=0)
                await self.state_store.joined(room_id, self.mxid)
            except MatrixRequestError as e2:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e2
        except MBadState as e:
            if not self.bot:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e
            try:
                await self.bot.unban_user(room_id, self.mxid)
                await self.bot.invite_user(room_id, self.mxid)
                await self.join_room(room_id, max_retries=0)
                await self.state_store.joined(room_id, self.mxid)
            except MatrixRequestError as e2:
                raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e2
        except MatrixRequestError as e:
            raise IntentError(f"Failed to join room {room_id} as {self.mxid}") from e

    def _register(self) -> Awaitable[dict]:
        content = {"username": self.localpart, "type": "m.login.application_service"}
        query_params = {"kind": "user"}
        return self.api.request(Method.POST, Path.register, content, query_params=query_params)

    async def ensure_registered(self) -> None:
        if await self.state_store.is_registered(self.mxid):
            return
        try:
            await self._register()
        except MatrixRequestError as e:
            if e.errcode != "M_USER_IN_USE":
                raise IntentError(f"Failed to register {self.mxid}", e)
                # self.log.exception(f"Failed to register {self.mxid}!")
                # return
        await self.state_store.registered(self.mxid)

    async def _ensure_has_power_level_for(self, room_id: RoomID, event_type: EventType) -> None:
        if not room_id:
            raise ValueError("Room ID not given")
        elif not event_type:
            raise ValueError("Event type not given")

        if not await self.state_store.has_power_levels_cached(room_id):
            await self.get_power_levels(room_id, ignore_cache=True)
        if not await self.state_store.has_power_level(room_id, self.mxid, event_type):
            # TODO implement something better
            raise IntentError(f"Power level of {self.mxid} is not enough "
                              f"for {event_type} in {room_id}")
            # self.log.warning(
            #     f"Power level of {self.mxid} is not enough for {event_type} in {room_id}")

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Union, Any, List, Awaitable, TYPE_CHECKING
import asyncio
import logging
import sys

import asyncpg

from .upgrade import UpgradeTable, upgrade_tables

if TYPE_CHECKING:
    from typing import Protocol


    class AcquireResult(Protocol):
        async def __aenter__(self) -> asyncpg.Connection: ...

        async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...

        def __await__(self) -> Awaitable[asyncpg.Connection]: ...


class Database:
    loop: asyncio.AbstractEventLoop
    log: logging.Logger

    _pool: Optional[asyncpg.pool.Pool]
    db_args: Dict[str, Any]
    upgrade_table: UpgradeTable

    url: str

    def __init__(self, url: str, db_args: Optional[Dict[str, Any]] = None,
                 upgrade_table: Union[None, UpgradeTable, str] = None,
                 log: Optional[logging.Logger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self.url = url
        self.db_args = db_args or {}
        if isinstance(upgrade_table, str):
            self.upgrade_table = upgrade_tables[upgrade_table]
        elif isinstance(upgrade_table, UpgradeTable):
            self.upgrade_table = upgrade_table
        elif upgrade_table is None:
            self.upgrade_table = UpgradeTable()
        else:
            raise ValueError(f"Can't use {type(upgrade_table)} as the upgrade table")
        self._pool = None
        self.log = log or logging.getLogger("mau.db")
        self.loop = loop or asyncio.get_event_loop()

    async def start(self) -> None:
        self.db_args["loop"] = self.loop
        self.log.debug(f"Connecting to {self.url}")
        self._pool = await asyncpg.create_pool(self.url, **self.db_args)
        try:
            await self.upgrade_table.upgrade(self.pool)
        except Exception:
            self.log.critical("Failed to upgrade database", exc_info=True)
            sys.exit(25)

    @property
    def pool(self) -> asyncpg.pool.Pool:
        if not self._pool:
            raise RuntimeError("Database has not been started")
        return self._pool

    async def stop(self) -> None:
        await self.pool.close()

    async def execute(self, query: str, *args: Any, timeout: Optional[float] = None) -> str:
        async with self.acquire() as conn:
            return await conn.execute(query, *args, timeout=timeout)

    async def fetch(self, query: str, *args: Any, timeout: Optional[float] = None
                    ) -> List[asyncpg.Record]:
        async with self.acquire() as conn:
            return await conn.fetch(query, *args, timeout=timeout)

    async def fetchval(self, query: str, *args: Any, column: int = 0,
                       timeout: Optional[float] = None) -> Any:
        async with self.acquire() as conn:
            return await conn.fetchval(query, *args, column=column, timeout=timeout)

    async def fetchrow(self, query: str, *args: Any, timeout: Optional[float] = None
                       ) -> asyncpg.Record:
        async with self.acquire() as conn:
            return await conn.fetchrow(query, *args, timeout=timeout)

    def acquire(self) -> 'AcquireResult':
        return self.pool.acquire()



from .as_handler import AppServiceServerMixin
from .appservice import AppService
from .api import AppServiceAPI, ChildAppServiceAPI, IntentAPI
from .state_store import ASStateStore

__all__ = ["AppService", "AppServiceAPI", "ChildAppServiceAPI", "IntentAPI", "ASStateStore",
           "AppServiceServerMixin"]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .type import EventType
from .base import GenericEvent
from .redaction import RedactionEventContent, RedactionEvent
from .message import (MessageEvent, MessageEventContent, MessageUnsigned, MediaMessageEventContent,
                      LocationMessageEventContent, LocationInfo, RelationType, MessageType, Format,
                      MediaInfo, FileInfo, AudioInfo, VideoInfo, ImageInfo, ThumbnailInfo,
                      TextMessageEventContent, BaseMessageEventContent, RelatesTo, BaseFileInfo,
                      EncryptedFile, JSONWebKey)
from .reaction import ReactionEventContent, ReactionEvent
from .state import (PowerLevelStateEventContent, Membership, MemberStateEventContent, StateEvent,
                    AliasesStateEventContent, CanonicalAliasStateEventContent, StrippedStateEvent,
                    RoomNameStateEventContent, RoomTopicStateEventContent, StateEventContent,
                    RoomPinnedEventsStateEventContent, StateUnsigned, RoomAvatarStateEventContent,
                    RoomTombstoneStateEventContent, RoomEncryptionStateEventContent)
from .account_data import (AccountDataEvent, AccountDataEventContent, RoomTagInfo,
                           RoomTagAccountDataEventContent)
from .ephemeral import (TypingEventContent, TypingEvent, PresenceEvent, PresenceState,
                        PresenceEventContent, SingleReceiptEventContent, ReceiptEventContent,
                        ReceiptEvent, ReceiptType, EphemeralEvent)
from .encrypted import (EncryptedEvent, EncryptedEventContent, EncryptionAlgorithm,
                        EncryptedOlmEventContent, EncryptedMegolmEventContent,
                        EncryptionKeyAlgorithm, OlmMsgType, OlmCiphertext)
from .to_device import (ToDeviceEvent, ToDeviceEventContent,  RoomKeyWithheldCode,
                        RoomKeyWithheldEventContent, RoomKeyEventContent, KeyRequestAction,
                        RequestedKeyInfo, RoomKeyRequestEventContent, ForwardedRoomKeyEventContent)
from .generic import Event, EventContent



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Type, cast
import logging

TRACE = logging.TRACE = 5
logging.addLevelName(TRACE, "TRACE")

SILLY = logging.SILLY = 1
logging.addLevelName(SILLY, "SILLY")

OldLogger: Type[logging.Logger] = cast(Type[logging.Logger], logging.getLoggerClass())


class TraceLogger(OldLogger):
    def trace(self, msg, *args, **kwargs) -> None:
        self.log(TRACE, msg, *args, **kwargs)

    def silly(self, msg, *args, **kwargs) -> None:
        self.log(SILLY, msg, *args, **kwargs)


logging.setLoggerClass(TraceLogger)



from .abstract import CryptoStore, StateStore
from .memory import MemoryCryptoStore
from .pickle import PickleCryptoStore
try:
    from .asyncpg import PgCryptoStore
except ImportError:
    PgCryptoStore = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Type
from abc import ABC, abstractmethod
import sys

from mautrix.types import RoomID, UserID
from mautrix.appservice import AppService, ASStateStore

from ..util.program import Program
from .config import BaseBridgeConfig
from .matrix import BaseMatrixHandler
from .portal import BasePortal
from .user import BaseUser
from .puppet import BasePuppet

try:
    from .state_store.sqlalchemy import SQLBridgeStateStore
    from ..util.db import Base

    import sqlalchemy as sql
    from sqlalchemy.engine.base import Engine
except ImportError:
    Base = SQLBridgeStateStore = sql = Engine = None

try:
    import uvloop
except ImportError:
    uvloop = None


class Bridge(Program, ABC):
    db: 'Engine'
    az: AppService
    state_store_class: Type[ASStateStore] = SQLBridgeStateStore
    state_store: ASStateStore
    config_class: Type[BaseBridgeConfig]
    config: BaseBridgeConfig
    matrix_class: Type[BaseMatrixHandler]
    matrix: BaseMatrixHandler
    repo_url: str
    markdown_version: str
    real_user_content_key: Optional[str] = None

    def __init__(self, module: str = None, name: str = None, description: str = None,
                 command: str = None, version: str = None,
                 real_user_content_key: Optional[str] = None,
                 config_class: Type[BaseBridgeConfig] = None,
                 matrix_class: Type[BaseMatrixHandler] = None,
                 state_store_class: Type[ASStateStore] = None) -> None:
        super().__init__(module, name, description, command, version, config_class)
        if real_user_content_key:
            self.real_user_content_key = real_user_content_key
        if matrix_class:
            self.matrix_class = matrix_class
        if state_store_class:
            self.state_store_class = state_store_class

    def prepare_arg_parser(self) -> None:
        super().prepare_arg_parser()
        self.parser.add_argument("-g", "--generate-registration", action="store_true",
                                 help="generate registration and quit")
        self.parser.add_argument("-r", "--registration", type=str, default="registration.yaml",
                                 metavar="<path>",
                                 help="the path to save the generated registration to (not needed "
                                      "for running the bridge)")

    def preinit(self) -> None:
        super().preinit()
        if self.args.generate_registration:
            self.generate_registration()
            sys.exit(0)

    def prepare(self) -> None:
        super().prepare()
        self.prepare_db()
        self.prepare_appservice()
        self.prepare_bridge()

    def prepare_config(self) -> None:
        self.config = self.config_class(self.args.config, self.args.registration,
                                        self.args.base_config)
        if self.args.generate_registration:
            self.config._check_tokens = False
        self.load_and_update_config()

    def generate_registration(self) -> None:
        self.config.generate_registration()
        self.config.save()
        print(f"Registration generated and saved to {self.config.registration_path}")

    def make_state_store(self) -> None:
        if self.state_store_class is None:
            raise RuntimeError("state_store_class is not set")
        elif SQLBridgeStateStore and issubclass(self.state_store_class, SQLBridgeStateStore):
            self.state_store = self.state_store_class(self.get_puppet, self.get_double_puppet)
        else:
            self.state_store = self.state_store_class()

    def prepare_appservice(self) -> None:
        self.make_state_store()
        mb = 1024 ** 2
        self.az = AppService(server=self.config["homeserver.address"],
                             domain=self.config["homeserver.domain"],
                             verify_ssl=self.config["homeserver.verify_ssl"],

                             as_token=self.config["appservice.as_token"],
                             hs_token=self.config["appservice.hs_token"],

                             tls_cert=self.config.get("appservice.tls_cert", None),
                             tls_key=self.config.get("appservice.tls_key", None),

                             bot_localpart=self.config["appservice.bot_username"],

                             log="mau.as",
                             loop=self.loop,

                             state_store=self.state_store,

                             real_user_content_key=self.real_user_content_key,

                             aiohttp_params={
                                 "client_max_size": self.config["appservice.max_body_size"] * mb
                             })

    def prepare_db(self) -> None:
        if not sql:
            raise RuntimeError("SQLAlchemy is not installed")
        self.db = sql.create_engine(self.config["appservice.database"])
        Base.metadata.bind = self.db
        if not self.db.has_table("alembic_version"):
            self.log.critical("alembic_version table not found. "
                              "Did you forget to `alembic upgrade head`?")
            sys.exit(10)

    def prepare_bridge(self) -> None:
        self.matrix = self.matrix_class(bridge=self)

    async def start(self) -> None:
        self.log.debug("Starting appservice...")
        await self.az.start(self.config["appservice.hostname"], self.config["appservice.port"])
        await self.matrix.wait_for_connection()
        await self.matrix.init_encryption()
        self.add_startup_actions(self.matrix.init_as_bot())
        await super().start()
        self.az.ready = True

    async def stop(self) -> None:
        await self.az.stop()
        await super().stop()
        if self.matrix.e2ee:
            await self.matrix.e2ee.stop()

    @abstractmethod
    async def get_user(self, user_id: UserID) -> 'BaseUser':
        pass

    @abstractmethod
    async def get_portal(self, room_id: RoomID) -> 'BasePortal':
        pass

    @abstractmethod
    async def get_puppet(self, user_id: UserID, create: bool = False) -> 'BasePuppet':
        pass

    @abstractmethod
    async def get_double_puppet(self, user_id: UserID) -> 'BasePuppet':
        pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from mautrix.errors import MatrixResponseError
from mautrix.api import Method, Path
from mautrix.types import UserSearchResults, Member, SerializerError, User

from .base import BaseClientAPI


class UserDataMethods(BaseClientAPI):
    """
    Methods in section 10 User Data of the spec. These methods are used for setting and getting user
    metadata and searching for users.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#user-data>`__
    """

    # region 10.1 User Directory
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#user-directory

    async def search_users(self, search_query: str, limit: Optional[int] = 10
                           ) -> UserSearchResults:
        """
        Performs a search for users on the homeserver. The homeserver may determine which subset of
        users are searched, however the homeserver MUST at a minimum consider the users the
        requesting user shares a room with and those who reside in public rooms (known to the
        homeserver). The search MUST consider local users to the homeserver, and SHOULD query remote
        users as part of the search.

        The search is performed case-insensitively on user IDs and display names preferably using a
        collation determined based upon the Accept-Language header provided in the request, if
        present.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#post-matrix-client-r0-user-directory-search>`__

        Args:
            search_query: The query to search for.
            limit: The maximum number of results to return.

        Returns:
            The results of the search and whether or not the results were limited.
        """
        content = await self.api.request(Method.POST, Path.user_directory.search, {
            "search_term": search_query,
            "limit": limit,
        })
        try:
            return UserSearchResults([User.deserialize(user) for user in content["results"]],
                                     content["limited"])
        except SerializerError as e:
            raise MatrixResponseError("Invalid user in search results") from e
        except KeyError:
            if "results" not in content:
                raise MatrixResponseError("`results` not in content.")
            elif "limited" not in content:
                raise MatrixResponseError("`limited` not in content.")
            raise

    # endregion
    # region 10.2 Profiles
    # API reference: https://matrix.org/docs/spec/client_server/r0.4.0.html#profiles

    async def set_displayname(self, displayname: str) -> None:
        """
        Set the display name of the current user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#put-matrix-client-r0-profile-userid-displayname>`__

        Args:
            displayname: The new display name for the user.
        """
        await self.api.request(Method.PUT, Path.profile[self.mxid].displayname, {
            "displayname": displayname,
        })

    async def get_displayname(self, user_id: str) -> str:
        """
        Get the display name of a user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#get-matrix-client-r0-profile-userid-displayname>`__

        Args:
            user_id: The ID of the user whose display name to get.

        Returns:
            The display name of the given user.
        """
        content = await self.api.request(Method.GET, Path.profile[user_id].displayname)
        try:
            return content["displayname"]
        except KeyError:
            raise MatrixResponseError("`displayname` not in response.")

    async def set_avatar_url(self, avatar_url: str) -> None:
        """
        Set the avatar of the current user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#put-matrix-client-r0-profile-userid-avatar-url>`__

        Args:
            avatar_url: The ``mxc://`` URI to the new avatar.
        """
        await self.api.request(Method.PUT, Path.profile[self.mxid].avatar_url, {
            "avatar_url": avatar_url,
        })

    async def get_avatar_url(self, user_id: str) -> str:
        """
        Get the avatar URL of a user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#get-matrix-client-r0-profile-userid-avatar-url>`__

        Args:
            user_id: The ID of the user whose avatar to get.

        Returns:
            The ``mxc://`` URI to the user's avatar.
        """
        content = await self.api.request(Method.GET, Path.profile[user_id].avatar_url)
        try:
            return content["avatar_url"]
        except KeyError:
            raise MatrixResponseError("`avatar_url` not in response.")

    async def get_profile(self, user_id: str) -> Member:
        """
        Get the combined profile information for a user.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#get-matrix-client-r0-profile-userid>`__

        Args:
            user_id: The ID of the user whose profile to get.

        Returns:
            The profile information of the given user.
        """
        content = await self.api.request(Method.GET, Path.profile[user_id])
        try:
            return Member.deserialize(content)
        except SerializerError as e:
            raise MatrixResponseError("Invalid member in response") from e

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, List, Dict, Union
from attr import dataclass
import attr

from ..primitive import JSON, UserID, EventID, ContentURI, RoomID, RoomAlias
from ..util import SerializableEnum, SerializableAttrs, Obj, deserializer
from .base import BaseRoomEvent, BaseUnsigned, EventType
from .encrypted import EncryptionAlgorithm


@dataclass
class PowerLevelStateEventContent(SerializableAttrs['PowerLevelStateEventContent']):
    """The content of a power level event."""
    users: Dict[UserID, int] = attr.ib(default=attr.Factory(dict), metadata={"omitempty": False})
    users_default: int = 0

    events: Dict[EventType, int] = attr.ib(default=attr.Factory(dict),
                                           metadata={"omitempty": False})
    events_default: int = 0

    state_default: int = 50

    invite: int = 50
    kick: int = 50
    ban: int = 50
    redact: int = 50

    def get_user_level(self, user_id: UserID) -> int:
        return self.users.get(user_id, self.users_default)

    def set_user_level(self, user_id: UserID, level: int) -> None:
        if level == self.users_default:
            del self.users[user_id]
        else:
            self.users[user_id] = level

    def ensure_user_level(self, user_id: UserID, level: int) -> bool:
        if self.get_user_level(user_id) != level:
            self.set_user_level(user_id, level)
            return True
        return False

    def get_event_level(self, event_type: EventType) -> int:
        return self.events.get(event_type,
                               self.state_default if event_type.is_state else self.events_default)

    def set_event_level(self, event_type: EventType, level: int) -> None:
        if level == self.state_default if event_type.is_state else self.events_default:
            del self.events[event_type]
        else:
            self.events[event_type] = level

    def ensure_event_level(self, event_type: EventType, level: int) -> bool:
        if self.get_event_level(event_type) != level:
            self.set_event_level(event_type, level)
            return True
        return False


class Membership(SerializableEnum):
    """A room membership state."""
    JOIN = "join"
    LEAVE = "leave"
    INVITE = "invite"
    BAN = "ban"
    KNOCK = "knock"


@dataclass
class MemberStateEventContent(SerializableAttrs['MemberStateEventContent']):
    """The content of a membership event."""
    membership: Membership = Membership.LEAVE
    avatar_url: str = None
    displayname: str = None
    is_direct: bool = False
    reason: str = None
    third_party_invite: JSON = None


@dataclass
class AliasesStateEventContent(SerializableAttrs['AliasesStateEventContent']):
    aliases: List[RoomAlias] = None


@dataclass
class CanonicalAliasStateEventContent(SerializableAttrs['CanonicalAliasStateEventContent']):
    canonical_alias: RoomAlias = attr.ib(default=None, metadata={"json": "alias"})
    alt_aliases: List[RoomAlias] = attr.ib(factory=lambda: [])


@dataclass
class RoomNameStateEventContent(SerializableAttrs['RoomNameStateEventContent']):
    name: str = None


@dataclass
class RoomTopicStateEventContent(SerializableAttrs['RoomTopicStateEventContent']):
    topic: str = None


@dataclass
class RoomAvatarStateEventContent(SerializableAttrs['RoomAvatarStateEventContent']):
    url: Optional[ContentURI] = None


@dataclass
class RoomPinnedEventsStateEventContent(SerializableAttrs['RoomPinnedEventsStateEventContent']):
    pinned: List[EventID] = None


@dataclass
class RoomTombstoneStateEventContent(SerializableAttrs['RoomID, UserID']):
    body: str = None
    replacement_room: RoomID = None


@dataclass
class RoomEncryptionStateEventContent(SerializableAttrs['RoomEncryptionStateEventContent']):
    algorithm: EncryptionAlgorithm = None
    rotation_period_ms: int = 604800000
    rotation_period_msgs: int = 100


StateEventContent = Union[PowerLevelStateEventContent, MemberStateEventContent,
                          AliasesStateEventContent, CanonicalAliasStateEventContent,
                          RoomNameStateEventContent, RoomAvatarStateEventContent,
                          RoomTopicStateEventContent, RoomPinnedEventsStateEventContent,
                          RoomTombstoneStateEventContent, RoomEncryptionStateEventContent, Obj]


@dataclass
class StrippedStateUnsigned(BaseUnsigned, SerializableAttrs['StrippedStateUnsigned']):
    """Unsigned information sent with state events."""
    prev_content: StateEventContent = None
    prev_sender: UserID = None
    replaces_state: EventID = None


@dataclass
class StrippedStateEvent(SerializableAttrs['StrippedStateEvent']):
    """Stripped state events included with some invite events."""
    content: StateEventContent = None
    room_id: RoomID = None
    sender: UserID = None
    type: EventType = None
    state_key: str = None

    unsigned: Optional[StrippedStateUnsigned] = None

    @property
    def prev_content(self) -> StateEventContent:
        if self.unsigned and self.unsigned.prev_content:
            return self.unsigned.prev_content
        return state_event_content_map.get(self.type, Obj)()

    @classmethod
    def deserialize(cls, data: JSON) -> 'StrippedStateEvent':
        try:
            event_type = EventType.find(data.get("type", None))
            data.get("content", {})["__mautrix_event_type"] = event_type
            data.get("unsigned", {}).get("prev_content", {})["__mautrix_event_type"] = event_type
        except ValueError:
            pass
        return super().deserialize(data)


@dataclass
class StateUnsigned(StrippedStateUnsigned, SerializableAttrs['StateUnsigned']):
    invite_room_state: Optional[List[StrippedStateEvent]] = None


state_event_content_map = {
    EventType.ROOM_POWER_LEVELS: PowerLevelStateEventContent,
    EventType.ROOM_MEMBER: MemberStateEventContent,
    EventType.ROOM_ALIASES: AliasesStateEventContent,
    EventType.ROOM_PINNED_EVENTS: RoomPinnedEventsStateEventContent,
    EventType.ROOM_CANONICAL_ALIAS: CanonicalAliasStateEventContent,
    EventType.ROOM_NAME: RoomNameStateEventContent,
    EventType.ROOM_AVATAR: RoomAvatarStateEventContent,
    EventType.ROOM_TOPIC: RoomTopicStateEventContent,
    EventType.ROOM_TOMBSTONE: RoomTombstoneStateEventContent,
    EventType.ROOM_ENCRYPTION: RoomEncryptionStateEventContent,
}


@dataclass
class StateEvent(BaseRoomEvent, SerializableAttrs['StateEvent']):
    """A room state event."""
    state_key: str
    content: StateEventContent
    unsigned: Optional[StateUnsigned] = None

    @property
    def prev_content(self) -> StateEventContent:
        if self.unsigned and self.unsigned.prev_content:
            return self.unsigned.prev_content
        return state_event_content_map.get(self.type, Obj)()

    @classmethod
    def deserialize(cls, data: JSON) -> 'StateEvent':
        try:
            event_type = EventType.find(data.get("type"))
            data.get("content", {})["__mautrix_event_type"] = event_type
            if "prev_content" in data and "prev_content" not in data.get("unsigned", {}):
                data.setdefault("unsigned", {})["prev_content"] = data["prev_content"]
            data.get("unsigned", {}).get("prev_content", {})["__mautrix_event_type"] = event_type
        except ValueError:
            return Obj(**data)
        return super().deserialize(data)

    @staticmethod
    @deserializer(StateEventContent)
    def deserialize_content(data: JSON) -> StateEventContent:
        evt_type = data.pop("__mautrix_event_type", None)
        content_type = state_event_content_map.get(evt_type, None)
        if not content_type:
            return Obj(**data)
        return content_type.deserialize(data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Any, Optional
from abc import ABC
import logging
import asyncio

from mautrix.appservice import AppService
from mautrix.types import UserID, EventType, Membership
from mautrix.errors import MNotFound
from mautrix.util.logging import TraceLogger

from .portal import BasePortal


class BaseUser(ABC):
    log: TraceLogger = logging.getLogger("mau.user")
    az: AppService
    loop: asyncio.AbstractEventLoop

    is_whitelisted: bool
    mxid: UserID
    command_status: Optional[Dict[str, Any]]

    async def is_logged_in(self) -> bool:
        return False

    async def is_in_portal(self, portal: BasePortal) -> bool:
        try:
            member_event = await portal.main_intent.get_state_event(
                portal.mxid, EventType.ROOM_MEMBER, self.mxid)
        except MNotFound:
            return False
        return member_event and member_event.membership in (Membership.JOIN, Membership.INVITE)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, List

from mautrix.types import SyncToken, IdentityKey, SessionID, RoomID, EventID, UserID, DeviceID
from mautrix.client.state_store import SyncStore
from mautrix.util.async_db import Database
from mautrix.util.logging import TraceLogger

from ... import (OlmAccount, Session, InboundGroupSession, OutboundGroupSession, TrustState,
                 DeviceIdentity)
from ..abstract import CryptoStore
from .upgrade import upgrade_table


class PgCryptoStore(CryptoStore, SyncStore):
    upgrade_table = upgrade_table

    db: Database
    account_id: str
    pickle_key: str
    log: TraceLogger

    _sync_token: Optional[SyncToken]
    _device_id: Optional[DeviceID]
    _account: Optional[OlmAccount]

    def __init__(self, account_id: str, pickle_key: str, db: Database) -> None:
        self.db = db
        self.account_id = account_id
        self.pickle_key = pickle_key

        self._sync_token = None
        self._device_id = ""
        self._account = None

    async def get_device_id(self) -> Optional[DeviceID]:
        device_id = await self.db.fetchval("SELECT device_id FROM crypto_account "
                                           "WHERE account_id=$1", self.account_id)
        self._device_id = device_id or self._device_id
        return self._device_id

    async def put_device_id(self, device_id: DeviceID) -> None:
        await self.db.fetchval("UPDATE crypto_account SET device_id=$1 WHERE account_id=$2",
                               device_id, self.account_id)
        self._device_id = device_id

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        self._sync_token = next_batch
        await self.db.execute("UPDATE crypto_account SET sync_token=$1 WHERE account_id=$2",
                              self._sync_token, self.account_id)

    async def get_next_batch(self) -> SyncToken:
        if self._sync_token is None:
            self._sync_token = await self.db.fetchval("SELECT sync_token FROM crypto_account"
                                                      " WHERE account_id=$1", self.account_id)
        return self._sync_token

    async def put_account(self, account: OlmAccount) -> None:
        self._account = account
        pickle = account.pickle(self.pickle_key)
        await self.db.execute("INSERT INTO crypto_account (account_id, device_id, shared, "
                              "sync_token, account) VALUES($1, $2, $3, $4, $5) "
                              "ON CONFLICT (account_id) DO UPDATE SET shared=$3, sync_token=$4,"
                              "                                       account=$5",
                              self.account_id, self._device_id, account.shared,
                              self._sync_token or "", pickle)

    async def get_account(self) -> OlmAccount:
        if self._account is None:
            row = await self.db.fetchrow("SELECT shared, account, device_id FROM crypto_account "
                                         "WHERE account_id=$1", self.account_id)
            if row is not None:
                self._account = OlmAccount.from_pickle(row["account"], passphrase=self.pickle_key,
                                                       shared=row["shared"])
        return self._account

    async def has_session(self, key: IdentityKey) -> bool:
        val = await self.db.fetchval("SELECT session_id FROM crypto_olm_session "
                                     "WHERE sender_key=$1 AND account_id=$2", key, self.account_id)
        return val is not None

    async def get_sessions(self, key: IdentityKey) -> List[Session]:
        rows = await self.db.fetch("SELECT session, created_at, last_used FROM crypto_olm_session "
                                   "WHERE sender_key=$1 AND account_id=$2 ORDER BY session_id",
                                   key, self.account_id)
        sessions = []
        for row in rows:
            sess = Session.from_pickle(row["session"], passphrase=self.pickle_key,
                                       creation_time=row["created_at"], use_time=row["last_used"])
            sessions.append(sess)
        return sessions

    async def get_latest_session(self, key: IdentityKey) -> Optional[Session]:
        row = await self.db.fetchrow("SELECT session, created_at, last_used FROM crypto_olm_session"
                                     " WHERE sender_key=$1 AND account_id=$2"
                                     " ORDER BY session DESC LIMIT 1", key, self.account_id)
        if row is None:
            return None
        return Session.from_pickle(row["session"], passphrase=self.pickle_key,
                                   creation_time=row["created_at"], use_time=row["last_used"])

    async def add_session(self, key: IdentityKey, session: Session) -> None:
        pickle = session.pickle(self.pickle_key)
        await self.db.execute("INSERT INTO crypto_olm_session (session_id, sender_key, session, "
                              "created_at, last_used, account_id) VALUES ($1, $2, $3, $4, $5, $6)",
                              session.id, key, pickle, session.creation_time, session.use_time,
                              self.account_id)

    async def update_session(self, key: IdentityKey, session: Session) -> None:
        pickle = session.pickle(self.pickle_key)
        await self.db.execute("UPDATE crypto_olm_session SET session=$1, last_used=$2 "
                              "WHERE session_id=$3 AND account_id=$4",
                              pickle, session.use_time, session.id, self.account_id)

    async def put_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID, session: InboundGroupSession) -> None:
        pickle = session.pickle(self.pickle_key)
        forwarding_chains = ",".join(session.forwarding_chain)
        await self.db.execute("INSERT INTO crypto_megolm_inbound_session (session_id, sender_key, "
                              "signing_key, room_id, session, forwarding_chains, account_id) "
                              "VALUES ($1, $2, $3, $4, $5, $6, $7)",
                              session_id, sender_key, session.signing_key, room_id, pickle,
                              forwarding_chains, self.account_id)

    async def get_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID) -> Optional[InboundGroupSession]:
        row = await self.db.fetchrow("SELECT signing_key, session, forwarding_chains "
                                     "FROM crypto_megolm_inbound_session WHERE room_id=$1"
                                     " AND sender_key=$2 AND session_id=$3 AND account_id=$4",
                                     room_id, sender_key, session_id, self.account_id)
        if row is None:
            return None
        return InboundGroupSession.from_pickle(row["session"], passphrase=self.pickle_key,
                                               signing_key=row["signing_key"],
                                               sender_key=sender_key, room_id=room_id,
                                               forwarding_chain=row["forwarding_chains"].split(","))

    async def has_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID) -> bool:
        count = await self.db.fetchval("SELECT COUNT(session) "
                                       "FROM crypto_megolm_inbound_session WHERE room_id=$1"
                                       " AND sender_key=$2 AND session_id=$3 AND account_id=$4",
                                       room_id, sender_key, session_id, self.account_id)
        return count > 0

    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        pickle = session.pickle(self.pickle_key)
        await self.db.execute("INSERT INTO crypto_megolm_outbound_session (room_id, session_id, "
                              "session, shared, max_messages, message_count, max_age, created_at, "
                              "last_used, account_id) "
                              "VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)"
                              "ON CONFLICT (account_id, room_id) DO UPDATE SET session_id=$2,"
                              "session=$3, shared=$4, max_messages=$5, message_count=$6, "
                              "max_age=$7, created_at=$8, last_used=$9",
                              session.room_id, session.id, pickle, session.shared,
                              session.max_messages, session.message_count, session.max_age,
                              session.creation_time, session.use_time, self.account_id)

    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        pickle = session.pickle(self.pickle_key)
        await self.db.execute("UPDATE crypto_megolm_outbound_session "
                              "SET session=$1, message_count=$2, last_used=$3 "
                              "WHERE room_id=$4 AND session_id=$5 AND account_id=$6",
                              pickle, session.message_count, session.use_time, session.room_id,
                              session.id, self.account_id)

    async def get_outbound_group_session(self, room_id: RoomID) -> Optional[OutboundGroupSession]:
        row = await self.db.fetchrow("SELECT room_id, session_id, session, shared, max_messages, "
                                     "message_count, max_age, created_at, last_used "
                                     "FROM crypto_megolm_outbound_session "
                                     "WHERE room_id=$1 AND account_id=$2",
                                     room_id, self.account_id)
        if row is None:
            return None
        return OutboundGroupSession.from_pickle(row["session"], passphrase=self.pickle_key,
                                                room_id=row["room_id"], shared=row["shared"],
                                                max_messages=row["max_messages"],
                                                message_count=row["message_count"],
                                                max_age=row["max_age"], use_time=row["last_used"],
                                                creation_time=row["created_at"])

    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        await self.db.execute("DELETE FROM crypto_megolm_outbound_session "
                              "WHERE room_id=$1 AND account_id=$2",
                              room_id, self.account_id)

    async def remove_outbound_group_sessions(self, rooms: List[RoomID]) -> None:
        await self.db.execute("DELETE FROM crypto_megolm_outbound_session "
                              "WHERE room_id=ANY($1) AND account_id=$2",
                              rooms, self.account_id)

    async def validate_message_index(self, sender_key: IdentityKey, session_id: SessionID,
                                     event_id: EventID, index: int, timestamp: int) -> bool:
        row = await self.db.fetchrow("SELECT event_id, timestamp FROM crypto_message_index "
                                     'WHERE sender_key=$1 AND session_id=$2 AND "index"=$3',
                                     sender_key, session_id, index)
        if row is None:
            await self.db.execute("INSERT INTO crypto_message_index (sender_key, session_id, index,"
                                  " event_id, timestamp) VALUES ($1, $2, $3, $4, $5)",
                                  sender_key, session_id, index, event_id, timestamp)
            return True
        return row["event_id"] == event_id and row["timestamp"] == timestamp

    async def get_devices(self, user_id: UserID) -> Optional[Dict[DeviceID, DeviceIdentity]]:
        tracked_user_id = await self.db.fetchval("SELECT user_id FROM crypto_tracked_user "
                                                 "WHERE user_id=$1", user_id)
        if tracked_user_id is None:
            return None
        rows = await self.db.fetch("SELECT device_id, identity_key, signing_key, trust, deleted, "
                                   "name FROM crypto_device WHERE user_id=$1", user_id)
        result = {}
        for row in rows:
            result[row["device_id"]] = DeviceIdentity(user_id=user_id, device_id=row["device_id"],
                                                      identity_key=row["identity_key"],
                                                      signing_key=row["signing_key"],
                                                      trust=TrustState(row["trust"]),
                                                      deleted=row["deleted"], name=row["name"])
        return result

    async def get_device(self, user_id: UserID, device_id: DeviceID) -> Optional[DeviceIdentity]:
        row = await self.db.fetchrow("SELECT identity_key, signing_key, trust, deleted, name "
                                     "FROM crypto_device WHERE user_id=$1 AND device_id=$2",
                                     user_id, device_id)
        if row is None:
            return None
        return DeviceIdentity(user_id=user_id, device_id=device_id, name=row["name"],
                              identity_key=row["identity_key"], signing_key=row["signing_key"],
                              trust=TrustState(row["trust"]), deleted=row["deleted"])

    async def put_devices(self, user_id: UserID, devices: Dict[DeviceID, DeviceIdentity]) -> None:
        data = [
            (user_id, device_id, identity.identity_key, identity.signing_key,
             identity.trust, identity.deleted, identity.name)
            for device_id, identity in devices.items()
        ]
        columns = ["user_id", "device_id", "identity_key", "signing_key",
                   "trust", "deleted", "name"]
        async with self.db.acquire() as conn, conn.transaction():
            await conn.execute("INSERT INTO crypto_tracked_user (user_id) VALUES ($1) "
                               "ON CONFLICT (user_id) DO NOTHING", user_id)
            await conn.execute("DELETE FROM crypto_device WHERE user_id=$1", user_id)
            await conn.copy_records_to_table("crypto_device", records=data, columns=columns)

    async def filter_tracked_users(self, users: List[UserID]) -> List[UserID]:
        rows = await self.db.fetch("SELECT user_id FROM crypto_tracked_user "
                                   "WHERE user_id = ANY($1)", users)
        return [row["user_id"] for row in rows]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Type, Callable

from .base import MatrixError


class MatrixRequestError(MatrixError):
    """An error that was returned by the homeserver."""
    http_status: int
    message: Optional[str]
    errcode: str


class MatrixUnknownRequestError(MatrixRequestError):
    """An unknown error type returned by the homeserver."""

    def __init__(self, http_status: int = 0, text: str = "", errcode: Optional[str] = None,
                 message: Optional[str] = None) -> None:
        super().__init__(f"{http_status}: {text}")
        self.http_status: int = http_status
        self.text: str = text
        self.errcode: Optional[str] = errcode
        self.message: Optional[str] = message


class MatrixStandardRequestError(MatrixRequestError):
    """A standard error type returned by the homeserver."""
    errcode: str = None

    def __init__(self, http_status: int, message: str = "") -> None:
        super().__init__(message)
        self.http_status: int = http_status
        self.message: str = message


MxSRE = Type[MatrixStandardRequestError]
ec_map: Dict[str, MxSRE] = {}


def standard_error(code: str) -> Callable[[MxSRE], MxSRE]:
    def decorator(cls: MxSRE) -> MxSRE:
        cls.errcode = code
        ec_map[code] = cls
        return cls

    return decorator


def make_request_error(http_status: int, text: str, errcode: str,
                       message: str) -> MatrixRequestError:
    """
    Determine the correct exception class for the error code and create an instance of that class
    with the given values.

    Args:
        http_status: The HTTP status code.
        text: The raw response text.
        errcode: The errcode field in the response JSON.
        message: The error field in the response JSON.
    """
    try:
        ec_class = ec_map[errcode]
        return ec_class(http_status, message)
    except KeyError:
        return MatrixUnknownRequestError(http_status, text, errcode, message)


# Standard error codes from https://matrix.org/docs/spec/client_server/r0.4.0.html#api-standards
# Additionally some combining superclasses for some of the error codes

@standard_error("M_FORBIDDEN")
class MForbidden(MatrixStandardRequestError):
    pass


@standard_error("M_USER_DEACTIVATED")
class MUserDeactivated(MForbidden):
    pass


class MatrixInvalidToken(MatrixStandardRequestError):
    pass


@standard_error("M_UNKNOWN_TOKEN")
class MUnknownToken(MatrixInvalidToken):
    pass


@standard_error("M_MISSING_TOKEN")
class MMissingToken(MatrixInvalidToken):
    pass


class MatrixBadRequest(MatrixStandardRequestError):
    pass


class MatrixBadContent(MatrixBadRequest):
    pass


@standard_error("M_BAD_JSON")
class MBadJSON(MatrixBadContent):
    pass


@standard_error("M_NOT_JSON")
class MNotJSON(MatrixBadContent):
    pass


@standard_error("M_NOT_FOUND")
class MNotFound(MatrixStandardRequestError):
    pass


@standard_error("M_LIMIT_EXCEEDED")
class MLimitExceeded(MatrixStandardRequestError):
    pass


@standard_error("M_UNKNOWN")
class MUnknown(MatrixStandardRequestError):
    pass


@standard_error("M_UNRECOGNIZED")
class MUnrecognized(MatrixStandardRequestError):
    pass


@standard_error("M_UNAUTHORIZED")
class MUnauthorized(MatrixStandardRequestError):
    pass


@standard_error("M_USER_IN_USE")
class MUserInUse(MatrixStandardRequestError):
    pass


@standard_error("M_INVALID_USERNAME")
class MInvalidUsername(MatrixStandardRequestError):
    pass


@standard_error("M_ROOM_IN_USE")
class MRoomInUse(MatrixStandardRequestError):
    pass


@standard_error("M_INVALID_ROOM_STATE")
class MInvalidRoomState(MatrixStandardRequestError):
    pass


# TODO THREEPID_ errors


@standard_error("M_UNSUPPORTED_ROOM_VERSION")
class MUnsupportedRoomVersion(MatrixStandardRequestError):
    pass


@standard_error("M_INCOMPATIBLE_ROOM_VERSION")
class MIncompatibleRoomVersion(MatrixStandardRequestError):
    pass


@standard_error("M_BAD_STATE")
class MBadState(MatrixStandardRequestError):
    pass


@standard_error("M_GUEST_ACCESS_FORBIDDEN")
class MGuestAccessForbidden(MatrixStandardRequestError):
    pass


@standard_error("M_CAPTCHA_NEEDED")
class MCaptchaNeeded(MatrixStandardRequestError):
    pass


@standard_error("M_CAPTCHA_INVALID")
class MCaptchaInvalid(MatrixStandardRequestError):
    pass


@standard_error("M_MISSING_PARAM")
class MMissingParam(MatrixBadRequest):
    pass


@standard_error("M_INVALID_PARAM")
class MInvalidParam(MatrixBadRequest):
    pass


@standard_error("M_TOO_LARGE")
class MTooLarge(MatrixBadRequest):
    pass


@standard_error("M_EXCLUSIVE")
class MExclusive(MatrixStandardRequestError):
    pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Callable, Union, Awaitable

from mautrix.types import UserID
from mautrix.appservice.state_store.asyncpg import PgASStateStore
from mautrix.util.async_db import Database

from ..puppet import BasePuppet

GetPuppetFunc = Union[Callable[[UserID], Awaitable[BasePuppet]],
                      Callable[[UserID, bool], Awaitable[BasePuppet]]]


class PgBridgeStateStore(PgASStateStore):
    def __init__(self, db: Database, get_puppet: 'GetPuppetFunc',
                 get_double_puppet: 'GetPuppetFunc') -> None:
        super().__init__(db)
        self.get_puppet = get_puppet
        self.get_double_puppet = get_double_puppet

    async def is_registered(self, user_id: UserID) -> bool:
        puppet = await self.get_puppet(user_id)
        if puppet:
            return puppet.is_registered
        custom_puppet = await self.get_double_puppet(user_id)
        if custom_puppet:
            return True
        return await super().is_registered(user_id)

    async def registered(self, user_id: UserID) -> None:
        puppet = await self.get_puppet(user_id, True)
        if puppet:
            puppet.is_registered = True
            await puppet.save()
        else:
            await super().registered(user_id)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Any, Iterable, Type, Tuple, Optional, cast

from ..primitive import JSON
from .serializable import Serializable


def _is_descriptor(obj):
    return (hasattr(obj, '__get__') or
            hasattr(obj, '__set__') or
            hasattr(obj, '__delete__'))


class ExtensibleEnumMeta(type):
    _by_value: Dict[Any, 'ExtensibleEnum']
    _by_key: Dict[str, 'ExtensibleEnum']

    def __new__(mcs: Type['ExtensibleEnumMeta'], name: str, bases: Tuple[Type, ...],
                classdict: Dict[str, Any]) -> Type['ExtensibleEnum']:
        create = [(key, val) for key, val in classdict.items()
                  if not key.startswith("_") and not _is_descriptor(val)]
        classdict = {key: val for key, val in classdict.items()
                     if key.startswith("_") or _is_descriptor(val)}
        enum_class = cast(Type['ExtensibleEnum'], super().__new__(mcs, name, bases, classdict))
        for key, val in create:
            ExtensibleEnum.__new__(enum_class, val).key = key
        return enum_class

    def __bool__(cls: Type['ExtensibleEnum']) -> bool:
        return True

    def __contains__(cls: Type['ExtensibleEnum'], value: Any) -> bool:
        if isinstance(value, cls):
            return value in cls._by_value.values()
        else:
            return value in cls._by_value.keys()

    def __getattr__(cls: Type['ExtensibleEnum'], name: Any) -> 'ExtensibleEnum':
        try:
            return cls._by_key[name]
        except KeyError:
            raise AttributeError(name) from None

    def __setattr__(cls: Type['ExtensibleEnum'], key: str, value: Any) -> None:
        if key.startswith("_"):
            return super().__setattr__(key, value)
        if not isinstance(value, cls):
            value = cls(value)
        value.key = key

    def __getitem__(cls: Type['ExtensibleEnum'], name: str) -> 'ExtensibleEnum':
        try:
            return cls._by_key[name]
        except KeyError:
            raise KeyError(name) from None

    def __setitem__(cls: Type['ExtensibleEnum'], key: str, value: Any) -> None:
        return cls.__setattr__(cls, key, value)

    def __iter__(cls: Type['ExtensibleEnum']) -> Iterable['ExtensibleEnum']:
        return cls._by_key.values().__iter__()

    def __len__(cls: Type['ExtensibleEnum']) -> int:
        return len(cls._by_key)

    def __repr__(cls: Type['ExtensibleEnum']) -> str:
        return f"<ExtensibleEnum {cls.__name__!r}>"


class ExtensibleEnum(Serializable, metaclass=ExtensibleEnumMeta):
    _by_value: Dict[Any, 'ExtensibleEnum'] = {}
    _by_key: Dict[str, 'ExtensibleEnum'] = {}

    _inited: bool = False
    _key: Optional[str]
    value: Any

    def __init__(self, value: Any) -> None:
        if self._inited:
            return
        self.value = value
        self._key = None
        self._inited = True

    def __new__(cls: Type['ExtensibleEnum'], value: Any) -> 'ExtensibleEnum':
        try:
            return cls._by_value[value]
        except KeyError as e:
            self = super().__new__(cls)
            self.__objclass__ = cls
            self.__init__(value)
            cls._by_value[value] = self
            return self

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        if self._key:
            return f"<{self.__class__.__name__}.{self._key}: {self.value!r}>"
        else:
            return f"{self.__class__.__name__}({self.value!r})"

    @property
    def key(self) -> str:
        return self._key

    @key.setter
    def key(self, key: str) -> None:
        self._key = key
        self._by_key[key] = self

    def serialize(self) -> JSON:
        return self.value

    @classmethod
    def deserialize(cls, raw: JSON) -> Any:
        return cls(raw)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Callable, Awaitable, List, Dict, Optional, Union
import logging

import asyncpg

from ..logging import TraceLogger

Upgrade = Callable[[asyncpg.Connection], Awaitable[None]]


class UnsupportedDatabaseVersion(Exception):
    pass


async def noop_upgrade(_: asyncpg.Connection) -> None:
    pass


class UpgradeTable:
    upgrades: List[Upgrade]
    allow_unsupported: bool
    database_name: str
    version_table_name: str
    log: TraceLogger

    def __init__(self, allow_unsupported: bool = False, version_table_name: str = "version",
                 database_name: str = "database",
                 log: Optional[Union[logging.Logger, TraceLogger]] = None) -> None:
        self.upgrades = [noop_upgrade]
        self.allow_unsupported = allow_unsupported
        self.version_table_name = version_table_name
        self.database_name = database_name
        self.log = log or logging.getLogger("mau.db.upgrade")

    def register(self, index: int = -1, description: str = "", _outer_fn: Optional[Upgrade] = None
                 ) -> Union[Upgrade, Callable[[Upgrade], Upgrade]]:
        if isinstance(index, str):
            description = index
            index = -1

        def actually_register(fn: Upgrade) -> Upgrade:
            fn.__mau_db_upgrade_description__ = description
            if index == -1 or index == len(self.upgrades):
                self.upgrades.append(fn)
            else:
                if len(self.upgrades) <= index:
                    self.upgrades += [noop_upgrade] * (index - len(self.upgrades) + 1)
                self.upgrades[index] = fn
            return fn

        return actually_register(_outer_fn) if _outer_fn else actually_register

    async def _save_version(self, conn: asyncpg.Connection, version: int) -> None:
        self.log.trace(f"Saving current version (v{version}) to database")
        await conn.execute(f"DELETE FROM {self.version_table_name}")
        await conn.execute(f"INSERT INTO {self.version_table_name} (version) VALUES ($1)",
                           version)

    async def upgrade(self, pool: asyncpg.pool.Pool) -> None:
        async with pool.acquire() as conn:
            await conn.execute(f"""CREATE TABLE IF NOT EXISTS {self.version_table_name} (
                version INTEGER PRIMARY KEY
            )""")
            row: asyncpg.Record = await conn.fetchrow("SELECT version FROM "
                                                      f"{self.version_table_name} LIMIT 1")
            version = row["version"] if row else 0

            if len(self.upgrades) < version:
                error = (f"Unsupported database version v{version} "
                         f"(latest known is v{len(self.upgrades) - 1})")
                if not self.allow_unsupported:
                    raise UnsupportedDatabaseVersion(error)
                else:
                    self.log.warning(error)
                    return
            elif len(self.upgrades) == version:
                self.log.debug(f"Database at v{version}, not upgrading")
                return

            for new_version in range(version + 1, len(self.upgrades)):
                upgrade = self.upgrades[new_version]
                desc = getattr(upgrade, "__mau_db_upgrade_description__", None)
                suffix = f": {desc}" if desc else ""
                self.log.debug(f"Upgrading {self.database_name} "
                               f"from v{version} to v{new_version}{suffix}")
                async with conn.transaction():
                    await upgrade(conn)
                    version = new_version
                    await self._save_version(conn, version)


upgrade_tables: Dict[str, UpgradeTable] = {}


def register_upgrade_table_parent_module(name: str) -> None:
    upgrade_tables[name] = UpgradeTable()


def _find_upgrade_table(fn: Upgrade) -> UpgradeTable:
    try:
        module = fn.__module__
    except AttributeError as e:
        raise ValueError("Registering upgrades without an UpgradeTable requires the function "
                         "to have the __module__ attribute.") from e
    parts = module.split(".")
    used_parts = []
    last_error = None
    for part in parts:
        used_parts.append(part)
        try:
            return upgrade_tables[".".join(used_parts)]
        except KeyError as e:
            last_error = e
    raise KeyError("Registering upgrades without an UpgradeTable requires you to register a parent "
                   "module with register_upgrade_table_parent_module first.") from last_error


def register_upgrade(index: int = -1, description: str = "") -> Callable[[Upgrade], Upgrade]:
    def actually_register(fn: Upgrade) -> Upgrade:
        return _find_upgrade_table(fn).register(index, description, fn)

    return actually_register



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
from abc import ABC
import pkg_resources
import logging

from yarl import URL
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap

from .base import BaseConfig
from .recursive_dict import RecursiveDict

yaml = YAML()
yaml.indent(4)
yaml.width = 200

log: logging.Logger = logging.getLogger("mau.util.config")


class BaseFileConfig(BaseConfig, ABC):
    def __init__(self, path: str, base_path: str) -> None:
        super().__init__()
        self._data = CommentedMap()
        self.path: str = path
        self.base_path: str = base_path

    def load(self) -> None:
        with open(self.path, 'r') as stream:
            self._data = yaml.load(stream)

    def load_base(self) -> Optional[RecursiveDict[CommentedMap]]:
        if self.base_path.startswith("pkg://"):
            url = URL(self.base_path)
            return RecursiveDict(yaml.load(pkg_resources.resource_stream(url.host, url.path)),
                                 CommentedMap)
        try:
            with open(self.base_path, 'r') as stream:
                return RecursiveDict(yaml.load(stream), CommentedMap)
        except OSError:
            pass
        return None

    def save(self) -> None:
        try:
            with open(self.path, 'w') as stream:
                yaml.dump(self._data, stream)
        except OSError:
            log.exception(f"Failed to overwrite the config in {self.path}")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .base import MatrixError


class CryptoError(MatrixError):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class EncryptionError(CryptoError):
    pass


class SessionShareError(CryptoError):
    pass


class DecryptionError(CryptoError):
    pass


class MatchingSessionDecryptionError(DecryptionError):
    pass


class DeviceValidationError(EncryptionError):
    pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List
import logging

from asyncpg import Connection

from mautrix.util.async_db.upgrade import UpgradeTable

upgrade_table = UpgradeTable(version_table_name="crypto_version", database_name="crypto store",
                             log=logging.getLogger("mau.crypto.db.upgrade"))


@upgrade_table.register(description="Initial revision")
async def upgrade_v1(conn: Connection) -> None:
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_account (
        device_id  VARCHAR(255) PRIMARY KEY,
        shared     BOOLEAN      NOT NULL,
        sync_token TEXT         NOT NULL,
        account    bytea        NOT NULL
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_message_index (
        sender_key CHAR(43),
        session_id CHAR(43),
        "index"    INTEGER,
        event_id   VARCHAR(255) NOT NULL,
        timestamp  BIGINT       NOT NULL,
        PRIMARY KEY (sender_key, session_id, "index")
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_tracked_user (
        user_id VARCHAR(255) PRIMARY KEY
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_device (
        user_id      VARCHAR(255),
        device_id    VARCHAR(255),
        identity_key CHAR(43)      NOT NULL,
        signing_key  CHAR(43)      NOT NULL,
        trust        SMALLINT      NOT NULL,
        deleted      BOOLEAN       NOT NULL,
        name         VARCHAR(255)  NOT NULL,
        PRIMARY KEY (user_id, device_id)
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_olm_session (
        session_id   CHAR(43)  PRIMARY KEY,
        sender_key   CHAR(43)  NOT NULL,
        session      bytea     NOT NULL,
        created_at   timestamp NOT NULL,
        last_used    timestamp NOT NULL
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_megolm_inbound_session (
        session_id   CHAR(43)     PRIMARY KEY,
        sender_key   CHAR(43)     NOT NULL,
        signing_key  CHAR(43)     NOT NULL,
        room_id      VARCHAR(255) NOT NULL,
        session      bytea        NOT NULL,
        forwarding_chains TEXT    NOT NULL
    )""")
    await conn.execute("""CREATE TABLE IF NOT EXISTS crypto_megolm_outbound_session (
        room_id       VARCHAR(255) PRIMARY KEY,
        session_id    CHAR(43)     NOT NULL UNIQUE,
        session       bytea        NOT NULL,
        shared        BOOLEAN      NOT NULL,
        max_messages  INTEGER      NOT NULL,
        message_count INTEGER      NOT NULL,
        max_age       INTERVAL     NOT NULL,
        created_at    timestamp    NOT NULL,
        last_used     timestamp    NOT NULL
    )""")


@upgrade_table.register(description="Add account_id primary key column")
async def upgrade_v2(conn: Connection) -> None:
    async def add_account_id_column(table: str, pkey_columns: List[str]) -> None:
        await conn.execute(f"ALTER TABLE {table} ADD COLUMN account_id VARCHAR(255)")
        await conn.execute(f"UPDATE {table} SET account_id=''")
        await conn.execute(f"ALTER TABLE {table} ALTER COLUMN account_id SET NOT NULL")
        await conn.execute(f"ALTER TABLE {table} DROP CONSTRAINT {table}_pkey")
        pkey_columns.append("account_id")
        pkey_columns_str = ", ".join(f'"{col}"' for col in pkey_columns)
        await conn.execute(f"ALTER TABLE {table} ADD CONSTRAINT {table}_pkey "
                           f"PRIMARY KEY ({pkey_columns_str})")

    await add_account_id_column("crypto_account", [])
    await add_account_id_column("crypto_olm_session", ["session_id"])
    await add_account_id_column("crypto_megolm_inbound_session", ["session_id"])
    await add_account_id_column("crypto_megolm_outbound_session", ["room_id"])



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Optional, Union

from mautrix.errors import MatrixResponseError
from mautrix.api import Method, Path
from mautrix.types import (UserID, LoginType, UserIdentifier, LoginResponse, LoginFlow,
                           MatrixUserIdentifier)

from .base import BaseClientAPI


class ClientAuthenticationMethods(BaseClientAPI):
    """
    Methods in section 5 Authentication of the spec. These methods are used for setting and getting user
    metadata and searching for users.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1.html#client-authentication>`__
    """

    # region 5.5 Login
    # API reference: https://matrix.org/docs/spec/client_server/r0.6.1.html#login

    async def get_login_flows(self) -> List[LoginFlow]:
        """
        Get login flows supported by the homeserver.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-login>`__

        Returns:
            The list of login flows that the homeserver supports.
        """
        resp = await self.api.request(Method.GET, Path.login)
        try:
            return [LoginFlow.deserialize(flow) for flow in resp["flows"]]
        except KeyError:
            raise MatrixResponseError("`flows` not in response.")

    async def login(self, identifier: Optional[Union[UserIdentifier, UserID]] = None,
                    login_type: LoginType = LoginType.PASSWORD, device_name: Optional[str] = None,
                    device_id: Optional[str] = None, password: Optional[str] = None,
                    store_access_token: bool = True, update_hs_url: bool = False, **kwargs: str
                    ) -> LoginResponse:
        """
        Authenticates the user, and issues an access token they can use to authorize themself in
        subsequent requests.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#post-matrix-client-r0-login>`__

        Args:
            login_type: The login type being used.
            identifier: Identification information for the user.
            device_name: A display name to assign to the newly-created device.
                Ignored if ``device_id`` correspnods to a known device.
            device_id: ID of the client device. If this does not correspond to a known client
                device, a new device will be created. The server will auto-generate a device_id
                if this is not specified.
            password: The user's password. Required when `type` is `m.login.password`.
            store_access_token: Whether or not mautrix-python should store the returned access token
                in this ClientAPI instance for future requests.
            update_hs_url: Whether or not mautrix-python should use the returned homeserver URL
                in this ClientAPI instance for future requests.
            **kwargs: Additional arguments for other login types.

        Returns:
            The login response.
        """
        if identifier is None or isinstance(identifier, str):
            identifier = MatrixUserIdentifier(identifier or self.mxid)
        if password is not None:
            kwargs["password"] = password
        if device_name is not None:
            kwargs["initial_device_display_name"] = device_name
        if device_id:
            kwargs["device_id"] = device_id
        elif self.device_id:
            kwargs["device_id"] = self.device_id
        resp = await self.api.request(Method.POST, Path.login, {
            "type": str(login_type),
            "identifier": identifier.serialize(),
            **kwargs,
        })
        resp_data = LoginResponse.deserialize(resp)
        if store_access_token:
            self.mxid = resp_data.user_id
            self.device_id = resp_data.device_id
            self.api.token = resp_data.access_token
        if update_hs_url:
            base_url = resp_data.well_known.homeserver.base_url
            if base_url and base_url != self.api.base_url:
                self.log.debug("Login response contained new base URL, switching from "
                               f"{self.api.base_url} to {base_url}")
                self.api.base_url = base_url.rstrip("/")
        return resp_data

    async def logout(self, clear_access_token: bool = True) -> None:
        """
        Invalidates an existing access token, so that it can no longer be used for authorization.
        The device associated with the access token is also deleted.
        `Device keys <https://matrix.org/docs/spec/client_server/latest#device-keys>`__ for the
        device are deleted alongside the device.

        See also: `API reference <https://matrix.org/docs/spec/client_server/latest#post-matrix-client-r0-logout>`__

        Args:
            clear_access_token: Whether or not mautrix-python should forget the stored access token.
        """
        await self.api.request(Method.POST, Path.logout)
        if clear_access_token:
            self.api.token = ""
            self.device_id = ""

    async def logout_all(self, clear_access_token: bool = True) -> None:
        """
        Invalidates all access tokens for a user, so that they can no longer be used for
        authorization. This includes the access token that made this request. All devices for the
        user are also deleted.
        `Device keys <https://matrix.org/docs/spec/client_server/latest#device-keys>`__ for the
        device are deleted alongside the device.

        This endpoint does not require UI (user-interactive) authorization because UI authorization
        is designed to protect against attacks where the someone gets hold of a single access token
        then takes over the account. This endpoint invalidates all access tokens for the user,
        including the token used in the request, and therefore the attacker is unable to take over
        the account in this way.

        See also: `API reference <https://matrix.org/docs/spec/client_server/latest#post-matrix-client-r0-logout-all>`__

        Args:
            clear_access_token: Whether or not mautrix-python should forget the stored access token.
        """
        await self.api.request(Method.POST, Path.logout.all)
        if clear_access_token:
            self.api.token = ""
            self.device_id = ""

    # endregion

    # TODO other sections

    # region 5.7 Current account information
    # API reference: https://matrix.org/docs/spec/client_server/r0.6.1.html#current-account-information

    async def whoami(self) -> UserID:
        """
        Get information about the current user.

        Returns:
            The user ID of the current user.
        """
        resp = await self.api.request(Method.GET, Path.account.whoami)
        try:
            return resp["user_id"]
        except KeyError:
            raise MatrixResponseError("`user_id` not in response.")

    # endregion



from .serializable import Serializable, SerializableEnum, SerializerError
from .serializable_attrs import SerializableAttrs, serializer, deserializer, no_value
from .obj import Obj, Lst
from .enum import ExtensibleEnum



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, TYPE_CHECKING
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from mautrix.types import SyncToken


class SyncStore(ABC):
    """SyncStore persists information used by /sync."""

    @abstractmethod
    async def put_next_batch(self, next_batch: 'SyncToken') -> None: ...

    @abstractmethod
    async def get_next_batch(self) -> 'SyncToken': ...


class MemorySyncStore(SyncStore):
    """MemorySyncStore is a :class:`SyncStore` implementation that stores the data in memory."""

    def __init__(self, next_batch: Optional['SyncToken'] = None) -> None:
        self._next_batch: Optional['SyncToken'] = next_batch

    async def put_next_batch(self, next_batch: 'SyncToken') -> None:
        self._next_batch = next_batch

    async def get_next_batch(self) -> 'SyncToken':
        return self._next_batch



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Optional, Dict, List
import asyncio

from mautrix.types import RoomID, UserID, EventID, EventType, StateEvent, StateEventContent, Member

from .api import ClientAPI
from .state_store import StateStore


class StoreUpdatingAPI(ClientAPI):
    state_store: Optional[StateStore]

    def __init__(self, *args, state_store: Optional[StateStore] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.state_store = state_store

    async def get_state(self, room_id: RoomID) -> List[StateEvent]:
        state = await super().get_state(room_id)
        if self.state_store:
            update_members = self.state_store.set_members(room_id, {
                evt.state_key: evt.content for evt in state
                if evt.type == EventType.ROOM_MEMBER
            })
            await asyncio.gather(update_members,
                                 *[self.state_store.update_state(evt) for evt in state
                                   if evt.type != EventType.ROOM_MEMBER])
        return state

    async def send_state_event(self, room_id: RoomID, event_type: EventType,
                               content: Union[StateEventContent, Dict],
                               state_key: Optional[str] = "", **kwargs) -> EventID:
        event_id = await super().send_state_event(room_id, event_type, content, state_key, **kwargs)
        if self.state_store:
            fake_event = StateEvent(type=event_type, room_id=room_id, event_id=event_id,
                                    sender=self.mxid, state_key=state_key, timestamp=0,
                                    content=content)
            await self.state_store.update_state(fake_event)
        return event_id

    async def get_state_event(self, room_id: RoomID, event_type: EventType,
                              state_key: Optional[str] = None) -> StateEventContent:
        event = await super().get_state_event(room_id, event_type, state_key)
        if self.state_store:
            fake_event = StateEvent(type=event_type, room_id=room_id, event_id=EventID(""),
                                    sender=UserID(""), state_key=state_key, timestamp=0,
                                    content=event)
            await self.state_store.update_state(fake_event)
        return event

    async def get_joined_members(self, room_id: RoomID) -> Dict[UserID, Member]:
        members = await super().get_joined_members(room_id)
        if self.state_store:
            await self.state_store.set_members(room_id, members, joined_only=True)
        return members

    async def get_members(self, room_id: RoomID) -> List[StateEvent]:
        member_events = await super().get_members(room_id)
        if self.state_store:
            await self.state_store.set_members(room_id, {evt.state_key: evt.content
                                                         for evt in member_events})
        return member_events
